<?php

namespace ApiQL;

use Doctrine\DBAL\Query\QueryBuilder;
use ApiQL\Language\AbstractLanguage;
use ApiQL\Language\ApiQueryBuilder;
use ApiQL\Language\LanguageFactory;
use ApiQL\Language\Specs\{
    Equals,
    NotEquals,
    LessThan,
    GreaterThan,
    LessThanOrEqual,
    GreaterThanOrEqual,
    Like,
    NotLike,
    In,
    NotIn,
    IsNull,
    IsNotNull
};

/**
 * Class ApiQL
 *
 * @package ApiQL
 */
class ApiQL extends AbstractLanguage
{
    public function __construct(array $expression, QueryBuilder $builder, array $specs = [])
    {
        $this->expression = $expression;
        self::$builder = new ApiQueryBuilder($builder);
        self::$specs = array_merge([
            "Equals" => Equals::class,
            "NotEquals" => NotEquals::class,
            "IsNull" => IsNull::class,
            "LessThan" => LessThan::class,
            "GreaterThan" => GreaterThan::class,
            "IsNotNull" => IsNotNull::class,
            "LessThanOrEqual" => LessThanOrEqual::class,
            "GreaterThanOrEqual" => GreaterThanOrEqual::class,
            "Like" => Like::class,
            "NotLike" => NotLike::class,
            "In" => In::class,
            "NotIn" => NotIn::class
        ], $specs);
    }

    /**
     * Execute specification root expression
     */
    public function execute()
    {
        $this->executeFields();
        $this->executeOffset();
        $this->executeLimit();
        $this->executeOrder();
        $this->executeWhere();
    }

    /**
     * Execute fields part of expression
     */
    private function executeFields()
    {
        $isProjection = false;
        if (array_key_exists('fields', $this->expression)) {
            $isProjection = true;
            $fieldsName = 'fields';
        } elseif (array_key_exists('select', $this->expression)) {
            $isProjection = true;
            $fieldsName = 'select';
        }
        if ($isProjection) {
            $fields = !is_array($this->expression[$fieldsName]) ? explode(',', $this->expression[$fieldsName]) :
                        $this->expression[$fieldsName];
            if (!empty($fields)) {
                foreach ($fields as $field) {
                    self::$builder->addSelect("\"" . $field . "\"");
                }
            } elseif (empty(self::$builder->getQueryPart("select"))) {
                self::$builder->addSelect('*');
            }
            unset($this->expression[$fieldsName]);
        } elseif (empty(self::$builder->getQueryPart("select"))) {
            self::$builder->addSelect('*');
        }
    }

    /**
     * Execute offset part of expression
     */
    private function executeOffset()
    {
        if (array_key_exists('offset', $this->expression)) {
            self::$builder->setFirstResult($this->expression['offset']);
            unset($this->expression['offset']);
        }
    }

    /**
     * Execute limit part of expression
     */
    private function executeLimit()
    {
        if (array_key_exists('limit', $this->expression)) {
            self::$builder->setMaxResults($this->expression['limit']);
            unset($this->expression['limit']);
        }
    }

    /**
     * Execute fields part of expression
     */
    private function executeOrder()
    {
        if (array_key_exists('order', $this->expression)) {
            $fields = !is_array($this->expression['order']) ? explode(',', $this->expression['order']) :
                        $this->expression['order'];
            if (!empty($fields)) {
                foreach ($fields as $field) {
                    $pair = explode(':', $field);
                    $fieldName = $pair[0];
                    $orderType = (count($pair) == 2) ? $pair[1] : null;
                    self::$builder->addOrderBy($fieldName, $orderType);
                }
            }
            unset($this->expression['order']);
        }
    }

    /**
     * Execute where part of expression
     */
    private function executeWhere()
    {
        if (!empty($this->expression)) {
            $expression = $this->expression;
            if (array_key_exists('where', $this->expression)) {
                $expression = $this->expression['where'];
            }
            if (array_key_exists(self::SPEC, $expression)) {
                $this->executeSpec($expression);
            } else {
                if (
                    array_key_exists(self::RELATION, $expression) ||
                    array_key_exists(self::LOGIC, $expression)
                ) {
                    $this->executeKeyWhere($expression);
                } else {
                    $this->executeNoKeyWhere($expression);
                }
            }
        }
    }

    /**
     * Execute explicit specification
     */
    private function executeSpec(array $expression)
    {
        $specName = $expression[self::SPEC];
        unset($expression[self::SPEC]);
        $this->expression = array_merge([self::SPEC => $specName], $expression);
        $interpreter = LanguageFactory::build($this->expression);
        self::$builder->andWhere($interpreter->execute());
    }

    /**
     * Execute where part of expression when operator is not set explicitly with rel or logic key
     */
    private function executeKeyWhere(array $expression)
    {
        $logic = self::getLogicOperator($expression);
        $rel = self::getRelationOperator($expression);
        if (array_key_exists(self::RELATION, $expression)) {
            unset($this->expression[self::RELATION]);
        }
        if (array_key_exists(self::LOGIC, $expression)) {
            unset($this->expression[self::LOGIC]);
        }
        if (!is_null($logic)) {
            $expression = [$logic => []];
            if (is_null($rel)) {
                $rel = "eq";
            }
            foreach ($this->expression as $key => $value) {
                $expression[$logic][] = [$rel => [$key => $value]];
            }
            $interpreters = array_map(function ($item) {
                return LanguageFactory::build($item);
            }, $expression[$logic]);
            self::$builder->andWhere($this->method($logic, ...$interpreters));
        } else {
            $expression = [$rel => $this->expression];
            $interpreter = LanguageFactory::build($expression);
            return self::$builder->andWhere($interpreter->execute());
        }
    }

    /**
     * Execute where part of expression when operator is not set explicitly with rel or logic key
     */
    private function executeNoKeyWhere(array $expression)
    {
        $operator = array_key_first($expression);
        $data_ = $expression[$operator];
        if (self::isLogicOperator($operator)) {
            $interpreters = array_map(function ($item) {
                return LanguageFactory::build($item);
            }, $data_);
            return self::$builder->andWhere($this->method($operator, ...$interpreters));
        } elseif (self::isRelationOperator($operator) || self::isSpecification($operator)) {
            $interpreter = LanguageFactory::build($expression);
            return self::$builder->andWhere($interpreter->execute());
        } elseif (count($this->expression) == 1) {
            $expression = ["eq" => $this->expression];
            $interpreter = LanguageFactory::build($expression);
            return self::$builder->andWhere($interpreter->execute());
        } else {
            $logic = "and";
            $rel = "eq";
            $expression = [$logic => []];
            foreach ($this->expression as $key => $value) {
                $expression[$logic][] = [$rel => [$key => $value]];
            }
            $interpreters = array_map(function ($item) {
                return LanguageFactory::build($item);
            }, $expression[$logic]);
            self::$builder->andWhere($this->method($logic, ...$interpreters));
        }
    }
}
