<?php

namespace ApiQL\Language;

use Doctrine\DBAL\Query\QueryBuilder;

/**
 * Class QueryBuilderDecorator - decorator of real query builders
 *
 * @package ApiQL\Language
 */
class ApiQueryBuilder
{
    use JsonField;

    /**
     * Query builder implementation
     *
     * @var QueryBuilder
     */
    private $builder;

    /**
     * Expression implementation
     *
     * @var mixed
     */
    private $expression;

    /**
     * Construct query builder
     *
     * @param QueryBuilder $builder - decorated query builder
     */
    public function __construct(QueryBuilder $builder)
    {
        $this->builder = $builder;
        $this->expression = $builder->expr();
    }

    /**
     * Execute logic AND expression with expression builder
     *
     * @param mixed $lhs - left-hand side expression builder
     * @param mixed $rhs - right-hand side expression builder
     */
    public function andX(...$expressions)
    {
        return $this->expression->andX(...$expressions);
    }

    /**
     * Execute logic OR expression with expression builder
     *
     * @param mixed $lhs - left-hand side expression builder
     * @param mixed $rhs - right-hand side expression builder
     */
    public function orX(...$expressions)
    {
        return $this->expression->orX(...$expressions);
    }

    /**
     * Execute logic EQUALITY expression
     *
     * @param string $field - expression field
     * @param mixed $value - expression value
     */
    public function eq(string $field, $value)
    {
        $fieldName = $this->getJsonFieldName($field, $value);
        $fieldValue = $this->getJsonFieldValue($field, $value);
        $this->builder->setParameter($fieldName, $this->getJsonPropertyValue($value));
        return $this->expression->eq($fieldValue, sprintf(":%s", $fieldName));
    }

    /**
     * Execute logic NOT EQUAL expression
     *
     * @param string $field - expression field
     * @param mixed $value - expression value
     */
    public function neq(string $field, $value)
    {
        $fieldName = $this->getJsonFieldName($field, $value);
        $fieldValue = $this->getJsonFieldValue($field, $value);
        $this->builder->setParameter($fieldName, $this->getJsonPropertyValue($value));
        return $this->expression->neq($fieldValue, sprintf(":%s", $fieldName));
    }

    /**
     * Execute logic LESS THAN expression
     *
     * @param string $field - expression field
     * @param mixed $value - expression value
     */
    public function lt(string $field, $value)
    {
        $fieldName = $this->getJsonFieldName($field, $value);
        $fieldValue = $this->getJsonFieldValue($field, $value);
        $this->builder->setParameter($fieldName, $this->getJsonPropertyValue($value));
        return $this->expression->lt($fieldValue, sprintf(":%s", $fieldName));
    }

    /**
     * Execute logic LESS THAN OR EQUAL expression
     *
     * @param string $field - expression field
     * @param mixed $value - expression value
     */
    public function lte(string $field, $value)
    {
        $fieldName = $this->getJsonFieldName($field, $value);
        $fieldValue = $this->getJsonFieldValue($field, $value);
        $this->builder->setParameter($fieldName, $this->getJsonPropertyValue($value));
        return $this->expression->lte($fieldValue, sprintf(":%s", $fieldName));
    }

    /**
     * Execute logic GREATER THAN expression
     *
     * @param string $field - expression field
     * @param mixed $value - expression value
     */
    public function gt(string $field, $value)
    {
        $fieldName = $this->getJsonFieldName($field, $value);
        $fieldValue = $this->getJsonFieldValue($field, $value);
        $this->builder->setParameter($fieldName, $this->getJsonPropertyValue($value));
        return $this->expression->gt($fieldValue, sprintf(":%s", $fieldName));
    }

    /**
     * Execute logic GREATER THAN OR EQUAL expression
     *
     * @param string $field - expression field
     * @param mixed $value - expression value
     */
    public function gte(string $field, $value)
    {
        $fieldName = $this->getJsonFieldName($field, $value);
        $fieldValue = $this->getJsonFieldValue($field, $value);
        $this->builder->setParameter($fieldName, $this->getJsonPropertyValue($value));
        return $this->expression->gte($fieldValue, sprintf(":%s", $fieldName));
    }

    /**
     * Execute logic IS NULL expression
     *
     * @param string $field - expression field
     */
    public function isNull(string $field)
    {
        $fieldValue = $this->getJsonFieldValue($field);
        return $this->expression->isNull($fieldValue);
    }

    /**
     * Execute logic IS NOT NULL expression
     *
     * @param string $field - expression field
     */
    public function isNotNull(string $field)
    {
        $fieldValue = $this->getJsonFieldValue($field);
        return $this->expression->isNotNull($fieldValue);
    }

    /**
     * Execute logic LIKE expression
     *
     * @param string $field - expression field
     * @param mixed $value - expression value
     */
    public function like(string $field, $value)
    {
        $fieldName = $this->getJsonFieldName($field, $value);
        $fieldValue = $this->getJsonFieldValue($field, $value);
        $this->builder->setParameter($fieldName, $this->getJsonPropertyValue($value));
        return $this->expression->like($fieldValue, sprintf(":%s", $fieldName));
    }

    /**
     * Execute logic NOT LIKE expression
     *
     * @param string $field - expression field
     * @param mixed $value - expression value
     */
    public function notLike(string $field, $value)
    {
        $fieldName = $this->getJsonFieldName($field, $value);
        $fieldValue = $this->getJsonFieldValue($field, $value);
        $this->builder->setParameter($fieldName, $this->getJsonPropertyValue($value));
        return $this->expression->notLike($fieldValue, sprintf(":%s", $fieldName));
    }

    /**
     * Execute logic IN expression
     *
     * @param string $field - expression field
     * @param mixed $initValue - expression value
     */
    public function in(string $field, $initValue)
    {
        $value = $this->getJsonPropertyValue($initValue);
        $value = is_string($value) ? explode(',', $value) : $value;
        $fieldValue = $this->getJsonFieldValue($field, $initValue, true);
        return $this->expression->in($fieldValue, $value);
    }

    /**
     * Execute logic NOT IN expression
     *
     * @param string $field - expression field
     * @param mixed $initValue - expression value
     */
    public function notIn(string $field, $initValue)
    {
        $value = $this->getJsonPropertyValue($initValue);
        $value = is_string($value) ? explode(',', $value) : $value;
        $fieldValue = $this->getJsonFieldValue($field, $initValue, true);
        return $this->expression->notIn($fieldValue, $value);
    }

    /**
     * Execute specification expression
     *
     * @param string $specName - specification name
     * @param mixed $data - specification data
     */
    public function spec(string $specName, $data)
    {
        if (is_array($data)) {
            $parameter = array_key_first($data);
            $value = $data[$parameter];
            $spec = $parameter === 0 ? new $specName($value) : $spec = new $specName($parameter, $value);
        } else {
            $spec = new $specName($data);
        }
        return $spec->match($this->builder);
    }

    /**
     * Add field to SELECT part
     *
     * @param string $field - field name
     */
    public function addSelect(string $field)
    {
        $this->builder->addSelect($field);
    }

    /**
     * Set limit part
     *
     * @param int $limit - limit
     */
    public function setMaxResults(int $limit)
    {
        $this->builder->setMaxResults($limit);
    }

    /**
     * Set offset part
     *
     * @param int $offset - offset
     */
    public function setFirstResult(int $offset)
    {
        $this->builder->setFirstResult($offset);
    }

    /**
     * Set ordering
     *
     * @param string $field - field
     * @param string $orderType - order type asc|desc
     */
    public function addOrderBy(string $field, string $orderType = null)
    {
        $this->builder->addOrderBy($field, $orderType);
    }

    /**
     * Set where
     *
     * @param mixed $where - where
     */
    public function andWhere($where)
    {
        $this->builder->andWhere($where);
    }

    /**
     * Return query part
     *
     * @param string $part - Query part
     *
     * @return array
     */
    public function getQueryPart(string $part): array
    {
        return $this->builder->getQueryPart($part);
    }
}
