<?php

namespace Tests\Integration;

use PHPUnit\Framework\TestCase;
use Doctrine\DBAL\DriverManager;
use ApiQL\ApiQL;

final class ApiTest extends TestCase
{
    protected static $cnx;
    
    public static function setUpBeforeClass(): void
    {
        self::$cnx = DriverManager::getConnection([
            'driver' => getenv("TEST_DB_DRIVER"),
            'host' => getenv("TEST_DB_HOST"),
            'dbname' => getenv("TEST_DB_NAME"),
            'user' => getenv("TEST_DB_USER"),
            'password' => getenv("TEST_DB_PASSWORD"),
            'port' => getenv("TEST_DB_PORT"),
        ]);

        $stmt = self::$cnx->prepare("CREATE TABLE IF NOT EXISTS interpreter_test(id serial, a int, b int, 
                                    c int, d varchar, e int, f jsonb)");
        $stmt->execute();

        $stmt = self::$cnx->prepare("DELETE FROM interpreter_test");
        $stmt->execute();

        $stmt = self::$cnx->prepare("INSERT INTO interpreter_test(a, b, c, d, e, f) VALUES " .
                                    "(5,5,5,'one',null,'[{\"id\":\"is_fixed\",\"value\":\"true\"}]')," .
                                    "(1,6,6,'none',null,'[{\"id\":\"num\",\"value\":\"10\"}]')," .
                                    "(8,8,3,'once',null,'[{\"id\":\"num\",\"value\":\"2\"}]')," .
                                    "(1,2,10,'three',null,'[{\"id\":\"is_fixed\",\"value\":\"true\"}]')," .
                                    "(2,7,7,'two',1,'[{\"id\":\"num\",\"value\":\"5\"}]')");
        $stmt->execute();
    }

    public static function tearDownAfterClass(): void
    {
        $stmt = self::$cnx->prepare("DELETE FROM interpreter_test");
        $stmt->execute();

        $stmt = self::$cnx->prepare("DROP TABLE interpreter_test");
        $stmt->execute();
    }

    public function testEquals()
    {
        //arrange
        $payload = ["where" => ["eq" => ["a" => 1]]];
        $builder = self::$cnx->createQueryBuilder();
        $api = new ApiQL($payload, $builder);
        //act
        $builder->select('*')->from('interpreter_test');
        $api->execute();
        $data = $builder->execute()->fetchAll();
        //assert
        $this->assertCount(
            2,
            $data
        );
        
        //arrange
        $payload = ["where" => ["eq" => ["b" => 5]]];
        $builder = self::$cnx->createQueryBuilder();
        $api = new ApiQL($payload, $builder);
        //act
        $builder->select('*')->from('interpreter_test');
        $api->execute();
        $data = $builder->execute()->fetchAll();
        //assert
        $this->assertCount(
            1,
            $data
        );
    }

    public function testNotEquals()
    {
        //arrange
        $payload = ["where" => ["neq" => ["c" => 7]]];
        $builder = self::$cnx->createQueryBuilder();
        $api = new ApiQL($payload, $builder);
        //act
        $builder->select('*')->from('interpreter_test');
        $api->execute();
        $data = $builder->execute()->fetchAll();
        //assert
        $this->assertCount(
            4,
            $data
        );
    }

    public function testLessThan()
    {
        //arrange
        $payload = ["where" => ["lt" => ["b" => 7]]];
        $builder = self::$cnx->createQueryBuilder();
        $api = new ApiQL($payload, $builder);
        //act
        $builder->select('*')->from('interpreter_test');
        $api->execute();
        $data = $builder->execute()->fetchAll();
        //assert
        $this->assertCount(
            3,
            $data
        );
    }

    public function testLessThanOrEqual()
    {
        //arrange
        $payload = ["where" => ["lte" => ["b" => 7]]];
        $builder = self::$cnx->createQueryBuilder();
        $api = new ApiQL($payload, $builder);
        //act
        $builder->select('*')->from('interpreter_test');
        $api->execute();
        $data = $builder->execute()->fetchAll();
        //assert
        $this->assertCount(
            4,
            $data
        );
    }

    public function testGreaterThan()
    {
        //arrange
        $payload = ["where" => ["gt" => ["c" => 3]]];
        $builder = self::$cnx->createQueryBuilder();
        $api = new ApiQL($payload, $builder);
        //act
        $builder->select('*')->from('interpreter_test');
        $api->execute();
        $data = $builder->execute()->fetchAll();
        //assert
        $this->assertCount(
            4,
            $data
        );
    }

    public function testGreaterThanOrEqual()
    {
        //arrange
        $payload = ["where" => ["gte" => ["c" => 7]]];
        $builder = self::$cnx->createQueryBuilder();
        $api = new ApiQL($payload, $builder);
        //act
        $builder->select('*')->from('interpreter_test');
        $api->execute();
        $data = $builder->execute()->fetchAll();
        //assert
        $this->assertCount(
            2,
            $data
        );
    }

    public function testLike()
    {
        //arrange
        $payload = ["where" => ["like" => ["d" => "one"]]];
        $builder = self::$cnx->createQueryBuilder();
        $api = new ApiQL($payload, $builder);
        //act
        $builder->select('*')->from('interpreter_test');
        $api->execute();
        $data = $builder->execute()->fetchAll();
        //assert
        $this->assertCount(
            1,
            $data
        );

        //arrange
        $payload = ["where" => ["like" => ["d" => "%one"]]];
        $builder = self::$cnx->createQueryBuilder();
        $api = new ApiQL($payload, $builder);
        //act
        $builder->select('*')->from('interpreter_test');
        $api->execute();
        $data = $builder->execute()->fetchAll();
        //assert
        $this->assertCount(
            2,
            $data
        );
    }

    public function testNotLike()
    {
        //arrange
        $payload = ["where" => ["not_like" => ["d" => "one"]]];
        $builder = self::$cnx->createQueryBuilder();
        $api = new ApiQL($payload, $builder);
        //act
        $builder->select('*')->from('interpreter_test');
        $api->execute();
        $data = $builder->execute()->fetchAll();
        //assert
        $this->assertCount(
            4,
            $data
        );

        //arrange
        $payload = ["where" => ["not_like" => ["d" => "%one"]]];
        $builder = self::$cnx->createQueryBuilder();
        $api = new ApiQL($payload, $builder);
        //act
        $builder->select('*')->from('interpreter_test');
        $api->execute();
        $data = $builder->execute()->fetchAll();
        //assert
        $this->assertCount(
            3,
            $data
        );
    }

    public function testIsNull()
    {
        //arrange
        $payload = ["where" => ["is_null" => ["e"]]];
        $builder = self::$cnx->createQueryBuilder();
        $api = new ApiQL($payload, $builder);
        //act
        $builder->select('*')->from('interpreter_test');
        $api->execute();
        $data = $builder->execute()->fetchAll();
        //assert
        $this->assertCount(
            4,
            $data
        );
    }

    public function testIsNotNull()
    {
        //arrange
        $payload = ["where" => ["is_not_null" => ["e"]]];
        $builder = self::$cnx->createQueryBuilder();
        $api = new ApiQL($payload, $builder);
        //act
        $builder->select('*')->from('interpreter_test');
        $api->execute();
        $data = $builder->execute()->fetchAll();
        //assert
        $this->assertCount(
            1,
            $data
        );
    }

    public function testIn()
    {
        //arrange
        $payload = ["where" => ["in" => ["a" => [1, 2]]]];
        $builder = self::$cnx->createQueryBuilder();
        $api = new ApiQL($payload, $builder);
        //act
        $builder->select('*')->from('interpreter_test');
        $api->execute();
        $data = $builder->execute()->fetchAll();
        //assert
        $this->assertCount(
            3,
            $data
        );
    }

    public function testNotIn()
    {
        //arrange
        $payload = ["where" => ["not_in" => ["b" => [7, 2, 5]]]];
        $builder = self::$cnx->createQueryBuilder();
        $api = new ApiQL($payload, $builder);
        //act
        $builder->select('*')->from('interpreter_test');
        $api->execute();
        $data = $builder->execute()->fetchAll();
        //assert
        $this->assertCount(
            2,
            $data
        );
    }

    public function testSimpleAnd()
    {
        //arrange
        $payload = ["where" => ["and" => [["eq" => ["a" => 1]], ["eq" => ["b" => 2]]]]];
        $builder = self::$cnx->createQueryBuilder();
        $api = new ApiQL($payload, $builder);
        //act
        $builder->select('*')->from('interpreter_test');
        $api->execute();
        $data = $builder->execute()->fetchAll();
        //assert
        $this->assertCount(
            1,
            $data
        );
    }

    public function testSimpleOr()
    {
        //arrange
        $payload = ["where" => ["or" => [["eq" => ["a" => 1]], ["eq" => ["b" => 8]]]]];
        $builder = self::$cnx->createQueryBuilder();
        $api = new ApiQL($payload, $builder);
        //act
        $builder->select('*')->from('interpreter_test');
        $api->execute();
        $data = $builder->execute()->fetchAll();
        //assert
        $this->assertCount(
            3,
            $data
        );
    }

    public function testAndOr()
    {
        //arrange WHERE a = 1 AND (b = 6 OR c = 10)
        $payload = ["where" => ["and" => [["eq" => ["a" => 1]],["or" => [["eq" => ["b" => 6]],["eq" =>
                   ["c" => 10]]]]]]];
        $builder = self::$cnx->createQueryBuilder();
        $api = new ApiQL($payload, $builder);
        //act
        $builder->select('*')->from('interpreter_test');
        $api->execute();
        $data = $builder->execute()->fetchAll();
        //assert
        $this->assertCount(
            2,
            $data
        );
    }

    public function testSpecOr()
    {
        //arrange WHERE Equals(a = 1) AND (b = 6 OR c = 10)
        $payload = ["where" => ["and" => [["spec" => "Equals", "data" => ["a" => 1]],
                   ["or" => [["eq" => ["b" => 6]],["eq" => ["c" => 10]]]]]]];
        $builder = self::$cnx->createQueryBuilder();
        $api = new ApiQL($payload, $builder);
        //act
        $builder->select('*')->from('interpreter_test');
        $api->execute();
        $data = $builder->execute()->fetchAll();
        //assert
        $this->assertCount(
            2,
            $data
        );
    }

    public function testMultiOr()
    {
        //arrange
        $payload = ["where" => ["or" => [["eq" => ["a" => 1]], ["eq" => ["b" => 8]], ["eq" => ["c" => 7]]]]];
        $builder = self::$cnx->createQueryBuilder();
        $api = new ApiQL($payload, $builder);
        //act
        $builder->select('*')->from('interpreter_test');
        $api->execute();
        $data = $builder->execute()->fetchAll();
        //assert
        $this->assertCount(
            4,
            $data
        );
    }

    public function testMultiAnd()
    {
        //arrange
        $payload = ["where" => ["and" => [["eq" => ["a" => 1]], ["eq" => ["b" => 6]], ["eq" => ["c" => 6]]]]];
        $builder = self::$cnx->createQueryBuilder();
        $api = new ApiQL($payload, $builder);
        //act
        $builder->select('*')->from('interpreter_test');
        $api->execute();
        $data = $builder->execute()->fetchAll();
        //assert
        $this->assertCount(
            1,
            $data
        );
    }

    public function testMultiAndWithSpec()
    {
        //arrange
        $payload = ["where" => ["and" => [["spec" => "Equals", "data" => ["a" => 1]], ["eq" => ["b" => 6]],
                       ["eq" => ["c" => 6]]]]];
        $builder = self::$cnx->createQueryBuilder();
        $api = new ApiQL($payload, $builder);
        //act
        $builder->select('*')->from('interpreter_test');
        $api->execute();
        $data = $builder->execute()->fetchAll();
        //assert
        $this->assertCount(
            1,
            $data
        );
    }

    public function testMultiAndWithSpecPlusNestedOr()
    {
        //arrange
        $payload = ["where" => ["and" => [["spec" => "Equals", "data" => ["a" => 1]], ["eq" => ["b" => 6]],
        ["or" => [  ["eq" => ["c" => 5]], ["eq" => ["c" => 3]]  ]  ]]]];
        $builder = self::$cnx->createQueryBuilder();
        $api = new ApiQL($payload, $builder);
        //act
        $builder->select('*')->from('interpreter_test');
        $api->execute();
        $data = $builder->execute()->fetchAll();
        //assert
        $this->assertCount(
            0,
            $data
        );

        //arrange
        $payload = ["where" => ["and" => [["spec" => "Equals", "data" => ["a" => 1]], ["eq" => ["b" => 6]],
        ["or" => [["eq" => ["c" => 5]],["spec" => "Equals", "data" => ["c" => 3]]]]]]];
        $builder = self::$cnx->createQueryBuilder();
        $api = new ApiQL($payload, $builder);
        //act
        $builder->select('*')->from('interpreter_test');
        $api->execute();
        $data = $builder->execute()->fetchAll();
        //assert
        $this->assertCount(
            0,
            $data
        );
    }
     
    public function testRepoWithSpec()
    {
        //arrange
        $payload = ["where" => ["spec" => "Equals", "data" => ["a" => 1]]];
        $repo = new ApiRepository(self::$cnx);
        //act
        $data = $repo->get($payload);
        //assert
        $this->assertCount(
            2,
            $data
        );

        //arrange
        $payload = ["where" => ["or" => [["spec" => "Equals", "data" => ["a" => 1]],
        ["spec" => "Equals", "data" => ["b" => 8]]]]];
        //act
        $data = $repo->get($payload);
        //assert
        $this->assertCount(
            3,
            $data
        );

        //arrange
        $payload = ["where" => ["or" => [["eq" => ["a" => 1]],["spec" => "Equals", "data" => ["b" => 8]]]]];
        //act
        $data = $repo->get($payload);
        //assert
        $this->assertCount(
            3,
            $data
        );

        //arrange
        $payload = [];
        //act
        $data = $repo->get($payload);
        //assert
        $this->assertCount(
            5,
            $data
        );
    }

    public function testEmptyResultset()
    {
        //arrange
        $payload = ["where" => ["eq" => ["b" => 1000]]];
        $builder = self::$cnx->createQueryBuilder();
        $api = new ApiQL($payload, $builder);
        //act
        $builder->select('*')->from('interpreter_test');
        $api->execute();
        $data = $builder->execute()->fetchAll(\PDO::FETCH_ASSOC);
        //assert
        $this->assertCount(
            0,
            $data
        );
    }

    public function testShortForm()
    {
        //arrange
        $payload = ["a" => 1, "rel" => "eq"];
        $builder = self::$cnx->createQueryBuilder();
        $api = new ApiQL($payload, $builder);
        //act
        $builder->select('*')->from('interpreter_test');
        $api->execute();
        $data = $builder->execute()->fetchAll(\PDO::FETCH_ASSOC);
        //assert
        $this->assertCount(
            2,
            $data
        );

        //arrange
        $payload = ["a" => 1, "rel" => "neq"];
        $builder = self::$cnx->createQueryBuilder();
        $api = new ApiQL($payload, $builder);
        //act
        $builder->select('*')->from('interpreter_test');
        $api->execute();
        $data = $builder->execute()->fetchAll(\PDO::FETCH_ASSOC);
        //assert
        $this->assertCount(
            3,
            $data
        );

        //arrange
        $payload = ["eq" => ["a" => 1]];
        $builder = self::$cnx->createQueryBuilder();
        $api = new ApiQL($payload, $builder);
        //act
        $builder->select('*')->from('interpreter_test');
        $api->execute();
        $data = $builder->execute()->fetchAll(\PDO::FETCH_ASSOC);
        //assert
        $this->assertCount(
            2,
            $data
        );

        //arrange
        $payload = ["neq" => ["a" => 1]];
        $builder = self::$cnx->createQueryBuilder();
        $api = new ApiQL($payload, $builder);
        //act
        $builder->select('*')->from('interpreter_test');
        $api->execute();
        $data = $builder->execute()->fetchAll(\PDO::FETCH_ASSOC);
        //assert
        $this->assertCount(
            3,
            $data
        );
    }

    public function testSimpleGetQuery()
    {
        //arrange
        $payload = ["a" => 1];
        $builder = self::$cnx->createQueryBuilder();
        $api = new ApiQL($payload, $builder);
        //act
        $builder->select('*')->from('interpreter_test');
        $api->execute();
        $data = $builder->execute()->fetchAll(\PDO::FETCH_ASSOC);
        //assert
        $this->assertCount(
            2,
            $data
        );

        //arrange
        $payload = ["a" => 1, "b" => 6];
        $builder = self::$cnx->createQueryBuilder();
        $api = new ApiQL($payload, $builder);
        //act
        $builder->select('*')->from('interpreter_test');
        $api->execute();
        $data = $builder->execute()->fetchAll(\PDO::FETCH_ASSOC);
        //assert
        $this->assertCount(
            1,
            $data
        );

        /*
        //arrange
        $payload = ["a" => 1, "b" => 8, "logic" => "or"];
        $builder = self::$cnx->createQueryBuilder();
        $api = new ApiQL($payload, $builder);
        //act
        $builder->select('*')->from('interpreter_test');
        $api->execute();
        $data = $builder->execute()->fetchAll(\PDO::FETCH_ASSOC);
        //assert
        $this->assertCount(
            3,
            $data
        );

        //arrange
        $payload = ["a" => 1, "b" => 8, "logic" => "and", "rel" => "neq"];
        $builder = self::$cnx->createQueryBuilder();
        $api = new ApiQL($payload, $builder);
        //act
        $builder->select('*')->from('interpreter_test');
        $api->execute();
        $data = $builder->execute()->fetchAll(\PDO::FETCH_ASSOC);
        //assert
        $this->assertCount(
            2,
            $data
        );

        //arrange
        $payload = ["a" => [1, 5], "rel" => "in"];
        $builder = self::$cnx->createQueryBuilder();
        $api = new ApiQL($payload, $builder);
        //act
        $builder->select('*')->from('interpreter_test');
        $api->execute();
        $data = $builder->execute()->fetchAll(\PDO::FETCH_ASSOC);
        //assert
        $this->assertCount(
            3,
            $data
        );
        */
    }

    public function testRepoWithSimpleGet()
    {
        //arrange ?a=1&b=8
        $payload = ["a" => 1, "b" => 6];
        $repo = new ApiRepository(self::$cnx);
        //act
        $data = $repo->get($payload);
        //assert
        $this->assertCount(
            1,
            $data
        );
    }

    public function testSpecs()
    {
        //arrange
        $payload = ["where" => ["spec" => "NotEquals", "data" => ["a" => 1]]];
        $repo = new ApiRepository(self::$cnx);
        //act
        $data = $repo->get($payload);
        //assert
        $this->assertCount(
            3,
            $data
        );

        
        //arrange
        $payload = ["where" => ["spec" => "GreaterThan", "data" => ["b" => 5]]];
        $repo = new ApiRepository(self::$cnx);
        //act
        $data = $repo->get($payload);
        //assert
        $this->assertCount(
            3,
            $data
        );

        
        //arrange
        $payload = ["where" => ["spec" => "GreaterThanOrEqual", "data" => ["c" => 6]]];
        $repo = new ApiRepository(self::$cnx);
        //act
        $data = $repo->get($payload);
        //assert
        $this->assertCount(
            3,
            $data
        );

        //arrange
        $payload = ["where" => ["spec" => "LessThan", "data" => ["b" => 5]]];
        $repo = new ApiRepository(self::$cnx);
        //act
        $data = $repo->get($payload);
        //assert
        $this->assertCount(
            1,
            $data
        );

        //arrange
        $payload = ["where" => ["spec" => "LessThanOrEqual", "data" => ["c" => 6]]];
        $repo = new ApiRepository(self::$cnx);
        //act
        $data = $repo->get($payload);
        //assert
        $this->assertCount(
            3,
            $data
        );

        //arrange
        $payload = ["where" => ["spec" => "Like", "data" => ["d" => "%one"]]];
        $repo = new ApiRepository(self::$cnx);
        //act
        $data = $repo->get($payload);
        //assert
        $this->assertCount(
            2,
            $data
        );

        //arrange
        $payload = ["where" => ["spec" => "NotLike", "data" => ["d" => "one"]]];
        $repo = new ApiRepository(self::$cnx);
        //act
        $data = $repo->get($payload);
        //assert
        $this->assertCount(
            4,
            $data
        );
        
        //arrange
        $payload = ["where" => ["spec" => "IsNull", "field" => "e"]];
        $repo = new ApiRepository(self::$cnx);
        //act
        $data = $repo->get($payload);
        //assert
        $this->assertCount(
            4,
            $data
        );
       
        //arrange
        $payload = ["where" => ["spec" => "IsNotNull", "data" => ["e"]]];
        $repo = new ApiRepository(self::$cnx);
        //act
        $data = $repo->get($payload);
        //assert
        $this->assertCount(
            1,
            $data
        );
        
        //arrange
        $payload = ["spec" => "IsNull", "field" => "e"];
        $repo = new ApiRepository(self::$cnx);
        //act
        $data = $repo->get($payload);
        //assert
        $this->assertCount(
            4,
            $data
        );

        
        //arrange
        $payload = ["spec" => "In", "data" => ["a" => [1, 2]]];
        $repo = new ApiRepository(self::$cnx);
        //act
        $data = $repo->get($payload);
        //assert
        $this->assertCount(
            3,
            $data
        );

        //arrange
        $payload = ["spec" => "In", "a" => "1, 2"];
        $repo = new ApiRepository(self::$cnx);
        //act
        $data = $repo->get($payload);
        //assert
        $this->assertCount(
            3,
            $data
        );
        

        //arrange
        $payload = ["spec" => "in", "a" => "1,2"];
        $repo = new ApiRepository(self::$cnx);
        //act
        $data = $repo->get($payload);
        //assert
        $this->assertCount(
            3,
            $data
        );


        //arrange
        $payload = ["spec" => "NotIn", "data" => ["b" => [2, 5, 7]]];
        $repo = new ApiRepository(self::$cnx);
        //act
        $data = $repo->get($payload);
        //assert
        $this->assertCount(
            2,
            $data
        );

        //arrange
        $payload = ["spec" => "not_in", "data" => ["b" => [2, 5, 7]]];
        $repo = new ApiRepository(self::$cnx);
        //act
        $data = $repo->get($payload);
        //assert
        $this->assertCount(
            2,
            $data
        );

        //arrange
        $payload = ["spec" => "not_in", "data" => ["b" => "2,5,7"]];
        $repo = new ApiRepository(self::$cnx);
        //act
        $data = $repo->get($payload);
        //assert
        $this->assertCount(
            2,
            $data
        );
        
        //arrange
        $payload = ["spec" => "not_in", "b" => "2,5,7"];
        $repo = new ApiRepository(self::$cnx);
        //act
        $data = $repo->get($payload);
        //assert
        $this->assertCount(
            2,
            $data
        );
    }

    public function testDynamicFields()
    {
        //arrange
        $payload = ["spec" => "not_in", "b" => [2, 5, 7], "fields" => ["a", "c"]];
        $repo = new ApiRepository(self::$cnx);
        //act
        $data = $repo->get($payload);
        //assert
        $this->assertArrayHasKey(
            "a",
            $data[0]
        );
        //assert
        $this->assertArrayHasKey(
            "c",
            $data[0]
        );

        //arrange
        $payload = ["spec" => "not_in", "b" => [2, 5, 7], "select" => ["a", "c"]];
        $repo = new ApiRepository(self::$cnx);
        //act
        $data = $repo->get($payload);
        //assert
        $this->assertArrayHasKey(
            "a",
            $data[0]
        );
        //assert
        $this->assertArrayHasKey(
            "c",
            $data[0]
        );

        //arrange
        $payload = ["spec" => "not_in", "b" => "2,5,7", "fields" => ["a", "c"]];
        $repo = new ApiRepository(self::$cnx);
        //act
        $data = $repo->get($payload);
        //assert
        $this->assertArrayHasKey(
            "a",
            $data[0]
        );
        //assert
        $this->assertArrayHasKey(
            "c",
            $data[0]
        );

        //arrange ?spec=not_in&b=2,5,7&fields=a,c
        $payload = ["spec" => "not_in", "b" => "2,5,7", "fields" => "a,c"];
        $repo = new ApiRepository(self::$cnx);
        //act
        $data = $repo->get($payload);
        //assert
        $this->assertArrayHasKey(
            "a",
            $data[0]
        );
        //assert
        $this->assertArrayHasKey(
            "c",
            $data[0]
        );
    }

    public function testLimitOffset()
    {
        //arrange
        $payload = ["spec" => "in", "b" => [2, 5, 7], "fields" => ["b"], "limit" => 1];
        $repo = new ApiRepository(self::$cnx);
        //act
        $data = $repo->get($payload);
        //assert
        $this->assertCount(
            1,
            $data
        );

        //arrange ?spec=in&b=2,5,7&fields=b,c&limit=10&offset=1
        $payload = ["spec" => "in", "b" => "2,5,7", "fields" => "b,c", "limit" => 10, "offset" => 1];
        $repo = new ApiRepository(self::$cnx);
        //act
        $data = $repo->get($payload);
        //assert
        $this->assertCount(
            2,
            $data
        );
    }

    public function testPartsOnly()
    {
        //arrange ?spec=in&b=2,5,7&fields=b&order=b
        $payload = ["fields" => "a,b"];
        $repo = new ApiRepository(self::$cnx);
        //act
        $data = $repo->get($payload);
        //assert
        $this->assertCount(
            5,
            $data
        );
        $this->assertArrayHasKey(
            "a",
            $data[0]
        );
        $this->assertArrayHasKey(
            "b",
            $data[0]
        );
        $this->assertTrue(
            empty($data[0]["c"])
        );
        $this->assertTrue(
            empty($data[0]["d"])
        );
        $this->assertTrue(
            empty($data[0]["e"])
        );
        $this->assertTrue(
            empty($data[0]["f"])
        );

        $payload = ["order" => "b"];
        $repo = new ApiRepository(self::$cnx);
        //act
        $data = $repo->get($payload);
        $this->assertEquals(
            2,
            $data[0]["b"]
        );
        $this->assertEquals(
            8,
            $data[4]["b"]
        );

        //arrange
        $payload = ["order" => "b:desc"];
        $repo = new ApiRepository(self::$cnx);
        //act
        $data = $repo->get($payload);
        $this->assertEquals(
            8,
            $data[0]["b"]
        );
        $this->assertEquals(
            2,
            $data[4]["b"]
        );

        //arrange
        $payload = ["limit" => 1];
        $repo = new ApiRepository(self::$cnx);
        //act
        $data = $repo->get($payload);
        //assert
        $this->assertCount(
            1,
            $data
        );

        //arrange
        $payload = ["offset" => 1];
        $repo = new ApiRepository(self::$cnx);
        //act
        $data = $repo->get($payload);
        //assert
        $this->assertCount(
            4,
            $data
        );

        //arrange
        $payload = ["offset" => 2];
        $repo = new ApiRepository(self::$cnx);
        //act
        $data = $repo->get($payload);
        //assert
        $this->assertCount(
            3,
            $data
        );

        //arrange
        $payload = [];
        $repo = new ApiRepository(self::$cnx);
        //act
        $data = $repo->get($payload);
        //assert
        $this->assertCount(
            5,
            $data
        );
    }

    public function testOrderBy()
    {
        //arrange ?spec=in&b=2,5,7&fields=b&order=b
        $payload = ["spec" => "in", "b" => "2,5,7", "fields" => "b", "order" => "b"];
        $repo = new ApiRepository(self::$cnx);
        //act
        $data = $repo->get($payload);
        //assert
        $this->assertEquals(
            2,
            $data[0]["b"]
        );
        //arrange ?spec=in&b=2,5,7&fields=b&order=b
        $payload = ["spec" => "in", "b" => "2,5,7", "fields" => "b", "order" => "b:desc"];
        $repo = new ApiRepository(self::$cnx);
        //act
        $data = $repo->get($payload);
        //assert
        $this->assertEquals(
            7,
            $data[0]["b"]
        );

        //arrange ?a=1&b=8
        $payload = ["a" => 1, "fields" => "b", "order" => "b:asc"];
        $repo = new ApiRepository(self::$cnx);
        //act
        $data = $repo->get($payload);
        //assert
        $this->assertEquals(
            2,
            $data[0]["b"]
        );

        //arrange ?a=1&fields=b,c,d&order=b:desc
        $payload = ["a" => 1, "fields" => "b,c,d", "order" => "b:desc"];
        $repo = new ApiRepository(self::$cnx);
        //act
        $data = $repo->get($payload);
        //assert
        $this->assertEquals(
            6,
            $data[0]["b"]
        );

        //arrange
        $payload = ["a" => 1, "fields" => "b", "order" => "b:asc"];
        $repo = new ApiRepository(self::$cnx);
        //act
        $data = $repo->get($payload);
        //assert
        $this->assertEquals(
            2,
            $data[0]["b"]
        );

        //arrange
        $payload = ["a" => 1, "fields" => "b", "order" => "b:asc", "offset" => 1];
        $repo = new ApiRepository(self::$cnx);
        //act
        $data = $repo->get($payload);
        //assert
        $this->assertEquals(
            6,
            $data[0]["b"]
        );
    }

    public function testArrayFields()
    {
        
        //arrange
        $payload = ["f" => ["is_fixed" => "true"]];
        $repo = new ApiRepository(self::$cnx);
        //act
        $data = $repo->get($payload);
        //assert
        $this->assertCount(
            2,
            $data
        );

        //arrange
        $payload = ["f" => ["is_fixed" => true]];
        $repo = new ApiRepository(self::$cnx);
        //act
        $data = $repo->get($payload);
        //assert
        $this->assertCount(
            2,
            $data
        );

        //arrange
        $payload = ["f" => ["is_fixed" => "false"]];
        $repo = new ApiRepository(self::$cnx);
        //act
        $data = $repo->get($payload);
        //assert
        $this->assertCount(
            0,
            $data
        );

        //arrange
        $payload = ["f" => ["undefined" => "false"]];
        $repo = new ApiRepository(self::$cnx);
        //act
        $data = $repo->get($payload);
        //assert
        $this->assertCount(
            0,
            $data
        );

        //arrange
        $payload = ["f" => ["num" => 5], "rel" => "lt"];
        $repo = new ApiRepository(self::$cnx);
        //act
        $data = $repo->get($payload);
        $this->assertCount(
            1,
            $data
        );

        //arrange
        $payload = ["f" => ["num" => 5], "rel" => "gt"];
        $repo = new ApiRepository(self::$cnx);
        //act
        $data = $repo->get($payload);
        $this->assertCount(
            1,
            $data
        );

        //arrange
        $payload = ["f" => ["num" => 5.0], "rel" => "lte"];
        $repo = new ApiRepository(self::$cnx);
        //act
        $data = $repo->get($payload);
        $this->assertCount(
            2,
            $data
        );

        //arrange
        $payload = ["f" => ["num" => 5.0], "rel" => "gte"];
        $repo = new ApiRepository(self::$cnx);
        //act
        $data = $repo->get($payload);
        $this->assertCount(
            2,
            $data
        );

        
        //arrange
        $payload = ["f" => ["num" => [2, 5, 10]], "rel" => "in"];
        $repo = new ApiRepository(self::$cnx);
        //act
        $data = $repo->get($payload);
        $this->assertCount(
            3,
            $data
        );
        
        
        //arrange
        $payload = ["f" => ["num" => "2,5,10"], "rel" => "in"];
        $repo = new ApiRepository(self::$cnx);
        //act
        $data = $repo->get($payload);
        $this->assertCount(
            3,
            $data
        );

        
        //arrange
        $payload = ["f" => ["num" => "10"], "rel" => "not_in"];
        $repo = new ApiRepository(self::$cnx);
        //act
        $data = $repo->get($payload);
        $this->assertCount(
            2,
            $data
        );

        //arrange
        $payload = ["f" => ["num" => 10], "rel" => "not_in"];
        $repo = new ApiRepository(self::$cnx);
        //act
        $data = $repo->get($payload);
        $this->assertCount(
            2,
            $data
        );
    }

    public function testJsonFields()
    {
        //arrange
        $payload = ["f[is_fixed]" => 'true'];
        $repo = new ApiRepository(self::$cnx);
        //act
        $data = $repo->get($payload);
        //assert
        $this->assertCount(
            2,
            $data
        );
        
        //arrange
        $payload = ["f[is_fixed]" => true];
        $repo = new ApiRepository(self::$cnx);
        //act
        $data = $repo->get($payload);
        //assert
        $this->assertCount(
            2,
            $data
        );

        //arrange
        $payload = ["f[is_fixed]" => 'false'];
        $repo = new ApiRepository(self::$cnx);
        //act
        $data = $repo->get($payload);
        //assert
        $this->assertCount(
            0,
            $data
        );

        //arrange
        $payload = ["f[undefined]" => 'false'];
        $repo = new ApiRepository(self::$cnx);
        //act
        $data = $repo->get($payload);
        //assert
        $this->assertCount(
            0,
            $data
        );

        //arrange
        $payload = ["f[num]" => 5, "rel" => "lt"];
        $repo = new ApiRepository(self::$cnx);
        //act
        $data = $repo->get($payload);
        $this->assertCount(
            1,
            $data
        );

        //arrange
        $payload = ["f[num]" => 5, "rel" => "gt"];
        $repo = new ApiRepository(self::$cnx);
        //act
        $data = $repo->get($payload);
        $this->assertCount(
            1,
            $data
        );

        //arrange
        $payload = ["f[num]" => 5.0, "rel" => "lte"];
        $repo = new ApiRepository(self::$cnx);
        //act
        $data = $repo->get($payload);
        $this->assertCount(
            2,
            $data
        );

        //arrange
        $payload = ["f[num]" => 5.0, "rel" => "gte"];
        $repo = new ApiRepository(self::$cnx);
        //act
        $data = $repo->get($payload);
        $this->assertCount(
            2,
            $data
        );

        //arrange
        $payload = ["f[num]" => [2, 5, 10], "rel" => "in"];
        $repo = new ApiRepository(self::$cnx);
        //act
        $data = $repo->get($payload);
        $this->assertCount(
            3,
            $data
        );

        //arrange
        $payload = ["f[num]" => "2,5,10", "rel" => "in"];
        $repo = new ApiRepository(self::$cnx);
        //act
        $data = $repo->get($payload);
        $this->assertCount(
            3,
            $data
        );

        //arrange
        $payload = ["f[num]" => "10", "rel" => "not_in"];
        $repo = new ApiRepository(self::$cnx);
        //act
        $data = $repo->get($payload);
        $this->assertCount(
            2,
            $data
        );

        //arrange
        $payload = ["f[num]" => 10, "rel" => "not_in"];
        $repo = new ApiRepository(self::$cnx);
        //act
        $data = $repo->get($payload);
        $this->assertCount(
            2,
            $data
        );
    }

    public function testJsonFieldSpecs()
    {
        //arrange
        $payload = ["spec" => "Equals", "f[is_fixed]" => 'true'];
        $repo = new ApiRepository(self::$cnx);
        //act
        $data = $repo->get($payload);
        //assert
        $this->assertCount(
            2,
            $data
        );

        //arrange
        $payload = ["spec" => "Equals", "f" => ['is_fixed' => 'true']];
        $repo = new ApiRepository(self::$cnx);
        //act
        $data = $repo->get($payload);
        //assert
        $this->assertCount(
            2,
            $data
        );

        //arrange
        $payload = ["spec" => "In", "f[num]" => "2,5,10"];
        $repo = new ApiRepository(self::$cnx);
        //act
        $data = $repo->get($payload);
        $this->assertCount(
            3,
            $data
        );

        //arrange
        $payload = ["spec" => "In", "f" => ["num" => "2,5,10"]];
        $repo = new ApiRepository(self::$cnx);
        //act
        $data = $repo->get($payload);
        $this->assertCount(
            3,
            $data
        );

        //arrange
        $payload = ["f[num]" => "2,5,10", "spec" => "In"];
        $repo = new ApiRepository(self::$cnx);
        //act
        $data = $repo->get($payload);
        //assert
        $this->assertCount(
            3,
            $data
        );

        //arrange
        $payload = ["f" => ["num" => "2,5,10"], "spec" => "In"];
        $repo = new ApiRepository(self::$cnx);
        //act
        $data = $repo->get($payload);
        //assert
        $this->assertCount(
            3,
            $data
        );

        
        //arrange
        $payload = ["spec" => "NotIn", "f" => ["num" => "10"]];
        $repo = new ApiRepository(self::$cnx);
        //act
        $data = $repo->get($payload);
        //assert
        $this->assertCount(
            2,
            $data
        );

        //arrange
        $payload = ["spec" => "NotIn", "f[num]" => 10];
        $repo = new ApiRepository(self::$cnx);
        //act
        $data = $repo->get($payload);
        //assert
        $this->assertCount(
            2,
            $data
        );

        //arrange
        $payload = ["spec" => "NotIn", "f" => ["num" => 10]];
        $repo = new ApiRepository(self::$cnx);
        //act
        $data = $repo->get($payload);
        //assert
        $this->assertCount(
            2,
            $data
        );

        //arrange
        $payload = ["f[num]" => 10, "spec" => "NotIn"];
        $repo = new ApiRepository(self::$cnx);
        //act
        $data = $repo->get($payload);
        $this->assertCount(
            2,
            $data
        );

        //arrange
        $payload = ["f" => ["num" => 10], "spec" => "NotIn"];
        $repo = new ApiRepository(self::$cnx);
        //act
        $data = $repo->get($payload);
        //assert
        $this->assertCount(
            2,
            $data
        );

        //arrange
        $payload = ["f[num]" => 10, "spec" => "not_in"];
        $repo = new ApiRepository(self::$cnx);
        //act
        $data = $repo->get($payload);
        //assert
        $this->assertCount(
            2,
            $data
        );

        //arrange
        $payload = ["f" => ["num" => 10], "spec" => "not_in"];
        $repo = new ApiRepository(self::$cnx);
        //act
        $data = $repo->get($payload);
        //assert
        $this->assertCount(
            2,
            $data
        );
    }
}
